/*******************************************************************************
*                                                                              *
*    Leibniz Institute on Aging - Fritz Lipmann Institute                      *
*    Genomic Ananlysis                                                         *
*                                                                              *
*    niels jahn	                                                               *
*    2014,2015                                                                 *
*                                                                              *
*******************************************************************************/

#include "tag_primer_amplicon.h"
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <vector>
#include <boost/lexical_cast.hpp>
#include <map>
#include <iomanip>
#include <cmath>
#include "revcomp.h"

using namespace std;
using namespace boost;

void usage(char *s)
{	cout << s << ": Usage :" << endl;
	cout << "output written to current working directory" << endl;
	cout << "	-h	this help" << endl;
	cout << "	-1 fastq	R1 (forward)" << endl;
	cout << "	-2 fastq	R2 (reverse)" << endl;
	cout << "	-s		use single hits" << endl;
	cout << "			output written even if only 1 primer found" << endl;
	cout << "	-p tpa		tag_primer_amplicon file (multiple possible)" << endl;
	cout << "	-P tpa_list	list of tag_primer_amplicon files" << endl;
	cout << "	-T		print matching tag histogram" << endl;
	cout << "	-u		toupper fastq (usually not needed)" << endl;
	cout << "	-d number	acceptable size difference (default 3) (-1 means no len check)" << endl;
	cout << "	-a adaptor	cut adaptor from fastq" << endl;
	cout << "	-H		show length histogram" << endl;
	cout << "	-A dir		directory for amplicons" << endl;
	cout << "	-m max_len	maximal len possible (only used when -d is used" << endl;
	cout << "	-M		do NOT write to mismatch" << endl;
	cout << "	-i		use internal motif (if given in tpa)" << endl;
	cout << "			(not possible with -s (use single hits))" << endl;
	cout << "	-S		show size diff hist" << endl;
	cout << "	-C		cut sequences at required amplicon size +2*tag size" << endl;
	cout <<	"	-E		removes all seqs not having revcomp tag at the end" << endl;
	cout << "	-B		do not write buckets" << endl;
	cout << "	-W		set file number width (default 3)" << endl;
	cout << endl;
}

bool check_end_tag(const string &s1,const string &s2,const string &tag,bool use_single_hit)
{	string	rv_tag(revcomp(tag));
	if (!use_single_hit)
	{	if (s1.substr(s1.size()-rv_tag.size())!=rv_tag) return false;
		if (s2.substr(s2.size()-rv_tag.size())!=rv_tag) return false;
		return true;
	}
	else
	{	if (s1.substr(0,tag.size())==tag && s1.substr(s1.size()-rv_tag.size())==rv_tag)
		{	return true;
		}
		if (s2.substr(0,tag.size())==tag && s2.substr(s2.size()-rv_tag.size())==rv_tag)
		{	return true;
		}
		return false;
	}
}	

int main(int argc,char **argv)
{	// print command
	for (int i=0; i<argc; ++i)
	{	cout << argv[i] << ' ';
	}
	cout << endl << endl;
	string	fq_1,fq_2;
	
	all_tag_primer_amplicon	tpa;
	bool	print_matching_tag_histogram=false;
	int tag_size=3;
	bool	toupper_input=false;
	int	acceptable_diff=3;
	string	adaptor;
	bool	show_histogram=false;
	string	amplicon_dir;
	int 	max_len=-1;
	bool	dont_write_mismatch=false;
	bool	search_internal_motif=false;
	bool	show_size_diff_hist=false;
	bool	do_hard_cut=false;
	bool	do_check_end_tag=false;
	bool	use_single_hits=false;
	bool	dont_write_buckets=false;

	int	file_number_width=3;

	string	tpa_list;

	int k;
	while (-1!=(k=getopt(argc,argv,"h1:2:p:Tua:HA:P:d:m:MiSCEsBW")))
	{	switch (k)
		{	case 'h' :
				usage(argv[0]);
				return 0;
			case '1' :
				fq_1=string(optarg);
				break;
			case '2' :
				fq_2=string(optarg);
				break;
			case 'p' :
				tpa.add(optarg);
				break;
			case 'P' :	
				tpa_list=string(optarg);
				break;
			case 'T' :
				print_matching_tag_histogram=true;
				break;
			case 'u' :
				toupper_input=true;
				break;
			case 'a' :
				adaptor=string(optarg);
				break;
			case 'H' :
				show_histogram=true;
				break;
			case 'A' :
				amplicon_dir=string(optarg);
				break;
			case 'd' :
				acceptable_diff=lexical_cast<int>(optarg);
				break;
			case 'm' :
				max_len=lexical_cast<int>(optarg);
				break;
			case 'M' :
				dont_write_mismatch=true;
				break;
			case 'i' :
				search_internal_motif=true;
				break;
			case 'S' :
				show_size_diff_hist=true;
				break;
			case 'C' :
				do_hard_cut=true;
				break;
			case 'E' :
				do_check_end_tag=true;
				break;
			case 's' :
				use_single_hits=true;
				break;
			case 'B' :
				dont_write_buckets=true;
				break;
			case 'W' :
				file_number_width=lexical_cast<int>(optarg);
				break;
			default :
				cerr << "unhandled parameter '" << char(k) << "'" << endl;
				return 1;
		}
	}

	if (tpa_list!="")
	{	ifstream	in(tpa_list.c_str());
		if (!in || !in.is_open())
		{	cerr << "can't open '" << tpa_list << "'" << endl;
			return 5;
		}
		string path(tpa_list);
		if (path.find_last_of('/')!=string::npos)
		{	path.erase(path.find_last_of('/'));
		}
		else
		{	path=".";
		}
		while (true)
		{	string line;
			getline(in,line);
			if (in.eof())
			{	break;
			}
			tpa.add(path+"/"+line);
		}
	}	

	if (fq_1=="" || fq_2=="" || tpa.size()==0)
	{	cerr << "(some) input missing :" << endl;
		if (fq_1=="") cerr << "\tR1 missing" << endl;
		if (fq_2=="") cerr << "\tR2 missing" << endl;
		if (tpa.size()==0) cerr << "\tno tags" << endl;
		usage(argv[0]);
		return 1;
	}	
	
	tpa.toupper();
	if (!tpa.check_tag_size(tag_size))
	{	cerr << "tag sizes differing from " << tag_size << endl;
		return 4;
	}	

	if (amplicon_dir!="")
	{	tpa.write_amplicons(amplicon_dir);
	}

	string fq_1_out(fq_1);
	if (fq_1_out.find_last_of("/")!=string::npos)
	{	fq_1_out.erase(0,1+fq_1_out.find_last_of("/"));
	}
	string fq_2_out(fq_2);
	if (fq_2_out.find_last_of("/")!=string::npos)
	{	fq_2_out.erase(0,1+fq_2_out.find_last_of("/"));
	}

	map<int,ofstream*>	outputs_1;
	map<int,ofstream*>	outputs_2;
	if (!dont_write_mismatch)
	{	outputs_1[0]=new ofstream(string(fq_1_out+".missed").c_str());
		outputs_2[0]=new ofstream(string(fq_2_out+".missed").c_str());
	}

	for (all_tag_primer_amplicon::iterator i=tpa.begin(); i!=tpa.end(); ++i)
	{	ostringstream o;
		o << ".bucket_" << setfill('0') << setw(file_number_width) << i->get_bucket();
		if (!dont_write_buckets)
		{	outputs_1[i->get_bucket()]=new ofstream(string(fq_1_out+o.str()).c_str());
			outputs_2[i->get_bucket()]=new ofstream(string(fq_2_out+o.str()).c_str());
		}
	}	

	ifstream	in_1(fq_1.c_str());
	if (!in_1 || !in_1.is_open())
	{	cerr << "can't open '" << fq_1 << "'" << endl;
		return 2;
	}
	ifstream	in_2(fq_2.c_str());
	if (!in_2 || !in_2.is_open())
	{	cerr << "can't open '" << fq_2 << "'" << endl;
		return 3;
	}	

	vector<string>	lines_1(4);
	vector<string>	lines_2(4);

	int nr_entry=0;
	int missed=0;
	int hits=0;
	int size_mismatch=0;
	int adaptors_found=0;
	int internal_motif_missing=0;
	int end_tag_check_failed=0;

	map<string,int> th;

	map<int,int>	size_hist_1;
	map<int,int>	size_hist_2;

	map<pair<int,int>,int>	size_diff_hist;

	while (true)
	{	if (nr_entry%100000==0) 
		{	cout << '.' << flush;
		}
		for (int i=0; i<4; ++i)
		{	getline(in_1,lines_1[i]);
			getline(in_2,lines_2[i]);
		}	
		if (in_1.eof() || in_2.eof())
		{	if (!(in_1.eof() && in_2.eof()))
			{	cerr << "line count in '" << fq_1 << "' and '" << fq_2 << "' differs" << endl;
				return 5;
			}
			break;
		}	
		nr_entry++;
		if (toupper_input)
		{	for (size_t i=0; i<lines_1[1].size(); ++i) lines_1[1][i]=toupper(lines_1[1][i]);
			for (size_t i=0; i<lines_2[1].size(); ++i) lines_2[1][i]=toupper(lines_2[1][i]);
		}
		if (adaptor.size())
		{	size_t p=lines_1[1].find(adaptor);
			if (p!=string::npos)
			{	adaptors_found++;
				lines_1[1].erase(p);
				lines_1[3].erase(p);
			}
			p=lines_2[1].find(adaptor);
			if (p!=string::npos)
			{	adaptors_found++;
				lines_2[1].erase(p);
				lines_2[3].erase(p);
			}
		}	
		
		bool found=false;
		for (all_tag_primer_amplicon::iterator i=tpa.begin(); i!=tpa.end(); ++i)
		{	if (i->match_primers_iupac(lines_1[1],lines_2[1]) 
				|| (use_single_hits && i->match_primers_iupac_single(lines_1[1],lines_2[1]))
				)
			{	// check sizes	

				int req_size=i->amplicon.size()+2*i->tag.size();
				if (max_len!=-1 && req_size>max_len)
				{	req_size=max_len;
				}
size_hist_1[lines_1[1].size()]++;
size_hist_2[lines_2[1].size()]++;
			
				// internal motif
				bool internal_motif_found=false;
				if (search_internal_motif && i->internal_motif.size())
				{	string m(i->internal_motif);
					/*fasta_piece fp;
					fp.set_data(m);
					fp.revcomp();
					string m_revcomp(fp.get_data());*/
					string m_revcomp(revcomp(m));

					if ((lines_1[1].find(m)!=string::npos && lines_2[1].find(m_revcomp)!=string::npos)
						|| (lines_1[1].find(m_revcomp)!=string::npos && lines_2[1].find(m)!=string::npos))
					{	internal_motif_found=true;
					}
					else
					{	internal_motif_missing++;
					}
				}

				if (do_hard_cut)
				{	if ((int)lines_1[1].size()>req_size)
					{	lines_1[1].erase(req_size);
						lines_1[3].erase(req_size);
					}
					if ((int)lines_2[1].size()>req_size)
					{	lines_2[1].erase(req_size);
						lines_2[3].erase(req_size);
					}
				}	

				if (acceptable_diff<0 ||
					(
					abs((signed int)(lines_1[1].size())-req_size)<=acceptable_diff &&
					abs((signed int)(lines_2[1].size())-req_size)<=acceptable_diff
					))
				{	
					if (!search_internal_motif || (search_internal_motif && internal_motif_found) || (search_internal_motif && i->internal_motif.size()==0) )
					{	
						found=true;
						if (!do_check_end_tag || check_end_tag(lines_1[1],lines_2[1],i->tag.get_data(),use_single_hits))
						{	hits++;
							i->hits++;
							for (int j=0; j<4; ++j)
							{	if (!dont_write_buckets)
								{	*outputs_1[i->get_bucket()] << lines_1[j] << endl;
									*outputs_2[i->get_bucket()] << lines_2[j] << endl;
								}
							}
						}
						else
						{	end_tag_check_failed++;
						}
					}
				}	
				else
				{	size_mismatch++;
					int s1=lines_1[1].size();
					int s2=lines_2[1].size();
					size_diff_hist[pair<int,int>(s1,s2)]++;
				}
			}
		}
		th[lines_1[1].substr(0,tag_size)]++;
		if (!found)
		{	missed++;
			if (!dont_write_mismatch)
			{	for (int i=0; i<4; ++i)
				{	*outputs_1[0] << lines_1[i] << endl;
					*outputs_2[0] << lines_2[i] << endl;
				}	
			}	
		}
	}
	cout << endl;
	cout << nr_entry << " entries" << endl;
	cout << size_mismatch << " size mismatch" << endl;
	cout << hits << " hits" << endl;
	if (adaptor.size()) cout << adaptors_found << " adpators found" << endl;	
	if (search_internal_motif) cout << internal_motif_missing << " internal motif missed" << endl;
	if (do_check_end_tag) cout << end_tag_check_failed << " end tag check failed" << endl;

	if (print_matching_tag_histogram)
	{	for (map<string,int>::iterator i=th.begin(); i!=th.end(); ++i)
		{	cout << i->first << '\t' << i->second << endl;
		}
	}

	for (all_tag_primer_amplicon::iterator i=tpa.begin(); i!=tpa.end(); ++i)
	{	cout << i->info() << endl;
	}

	if (!dont_write_buckets)
	{	for (map<int,ofstream*>::iterator i=outputs_1.begin(); i!=outputs_1.end(); ++i)
		{	i->second->close();
		}
		for (map<int,ofstream*>::iterator i=outputs_2.begin(); i!=outputs_2.end(); ++i)
		{	i->second->close();
		}
	}

	if (show_histogram)
	{	for (map<int,int>::iterator i=size_hist_1.begin(); i!=size_hist_1.end(); ++i)
		{	cout << i->first << '\t' << i->second << endl;
		}
		cout << endl;
		for (map<int,int>::iterator i=size_hist_2.begin(); i!=size_hist_2.end(); ++i)
		{	cout << i->first << '\t' << i->second << endl;
		}
	}
	if (show_size_diff_hist)
	{	for (map<pair<int,int>,int>::iterator i=size_diff_hist.begin(); i!=size_diff_hist.end(); ++i)
		{	cout << i->first.first << "\t" << i->first.second << "\t" << i->second << endl;
		}
	}	

	for (map<int,ofstream*>::iterator i=outputs_1.begin(); i!=outputs_1.end(); ++i)
	{	i->second->close();
		delete i->second;
	}
	for (map<int,ofstream*>::iterator i=outputs_2.begin(); i!=outputs_2.end(); ++i)
	{	i->second->close();
		delete i->second;
	}

	return 0;
}


