/*******************************************************************************
*                                                                              *
*    Leibniz Institute on Aging - Fritz Lipmann Institute                      *
*    Genomic Ananlysis                                                         *
*                                                                              *
*    niels jahn	                                                               *
*    2014,2015                                                                 *
*                                                                              *
*******************************************************************************/

#include "fasta.h"
#include <fstream>
#include <sstream>
#include <ctype.h>
#include <iostream>
#include <sstream>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

using namespace std;

static const string zcat("/usr/local/bin/zcat");
static const string rm("/bin/rm");

fasta_piece::fasta_piece()
:name(),data()
{	startchar='>';
	output_width=60;
}

fasta_piece::fasta_piece(const fasta_piece &p)
:name(p.name),data(p.data)
{	startchar=p.startchar;
	output_width=p.output_width;
}

void fasta_piece::set_startchar(unsigned char s)
{	startchar=s;
}

void fasta_piece::set_output_width(int x)
{	output_width=x;
}

bool fasta_piece::blastn_valid() const
{	return string::npos==data.find_first_not_of("acgtACGTnN");
}

void fasta_piece::toupper()
{	for (unsigned int pos=0; pos<data.size(); ++pos)
		data[pos]=std::toupper(data[pos]);
}

void fasta_piece::tolower()
{	for (unsigned int pos=0; pos<data.size(); ++pos)
		data[pos]=std::tolower(data[pos]);
}

void fasta_piece::reverse()
{	for (unsigned int p=0; p<data.size()/2; ++p)
	{	unsigned char h=data[p];
		data[p]=data[data.size()-1-p];
		data[data.size()-1-p]=h;
	}
}

void fasta_piece::complement()
{	ostringstream	err;
	for (unsigned int p=0; p<data.size(); ++p)
	{	switch (data[p])
		{	case 'a' :
				data[p]='t'; break;
			case 'A' :
				data[p]='T'; break;
			case 'c' :
				data[p]='g'; break;
			case 'C' :
				data[p]='G'; break;
			case 'g' :
				data[p]='c'; break;
			case 'G' :
				data[p]='C'; break;
			case 't' :
				data[p]='a'; break;
			case 'T' :
				data[p]='A'; break;
			case 'n' :
			case 'N' :
				break;
			case 'r' :
				data[p]='y'; break;
			case 'R' :
				data[p]='Y'; break;
			case 'y' :
				data[p]='r'; break;
			case 'Y' :
				data[p]='R'; break;
			case 'm' :
				data[p]='k'; break;
			case 'M' :
				data[p]='K'; break;
			default :
				err << "fasta_piece::complement() : unhandled nucleotide '" << data[p] << "'";
				if (!isprint(data[p]))
					err << "0x" << hex << int(data[p]); 
				err << " at pos " << p;
				throw bio_exception(err.str());
		}
	}
}

void fasta_piece::revcomp()
{	reverse();
	complement();
}

void	fasta_piece::set(int s,int e,char c)
{	if (s<0) s=0;
	if (unsigned(e)>data.size()-1) e=data.size()-1;
	data.replace(s,1+e-s,1+e-s,c); // change including pos e
}

list<int> fasta_piece::search(const string &s) const
{	list<int>	ret;
	string::size_type pos=0;
	while (string::npos!=(pos=data.find(s,pos)))
		ret.push_back(pos++);
	return ret;	
}		

string	fasta_piece::get_part(int start,int len)
{	if (unsigned(start+len)>data.size())
	{	ostringstream	o;
		o << "fasta_piece::get_part : accessing nonexistent data data.size=" << data.size() << " start=" << start << " len=" << len;
		throw bio_exception(o.str());
	}
	string	ret(data.substr(start,len));
	return ret;
}

double	fasta_piece::get_gc() const
{	int	at=0;
	int	gc=0;
	for (unsigned int l=0; l<data.size(); ++l)
	{	switch (data[l])
		{	case 'a' :
			case 'A' :
			case 't' :
			case 'T' :
				at++;
				break;
			case 'g' :
			case 'G' :
			case 'c' :
			case 'C' :
				gc++;
				break;
		}
	}
	return double(gc)/double(at+gc);
}

istream& operator >> (istream &i,fasta_piece &f)
{	f.name.clear();
	f.data.clear();
	const unsigned int BLOCK_SIZE=256;
	unsigned int	size=BLOCK_SIZE;
	f.data.reserve(size);
	unsigned int c;
	c=i.get();
	if (c!=f.startchar)
	{	i.unget();
		throw bio_exception("fasta_piece : start_char missing");
		return i;
	}	
	c=i.get();
	while (c!=0xa && c!=0xd && !i.eof())
	{	f.name+=c;
		c=i.get();
	}	
	while (!i.eof() && c!=f.startchar)
	{	if (c!=0xa && c!=0xd) f.data+=c;
		c=i.get();
		if (f.data.length()==size)
		{	size+=BLOCK_SIZE;
			f.data.reserve(size);
		}	
	}	
	if (c==f.startchar) i.unget();
	return i;
}

ostream& operator << (ostream &o,const fasta_piece &f)
{	o << f.startchar  << f.name << endl;
	unsigned int pos=0;
	while (pos<f.data.length())
	{	o << f.data.substr(pos,f.output_width) << endl;
		pos+=f.output_width;
	}	
	return o;
}	

fasta::fasta()
:data()
{	
}

fasta::fasta(const fasta_piece &fp)
:data()
{	data.push_back(fp);
}

fasta::fasta(const fasta &f)
:data(f.data)
{	
}

fasta::fasta(const string &fn)
:data()
{	load(fn);
}

void fasta::clear()
{	data.clear();
}

void fasta::add(const fasta_piece &fp)
{	data.push_back(fp);
}

void fasta::add_after(int pos,const fasta_piece &fp)
{	int	l=0;
	for (list<fasta_piece>::iterator i=data.begin(); i!=data.end(); ++i,++l)
		if (l==pos)
		{	data.insert(i,fp);
			break;
		}	
}

void fasta::erase(int pos)
{	int	l=0;
	for (list<fasta_piece>::iterator i=data.begin(); i!=data.end(); ++i,++l)
		if (l==pos)
		{	data.erase(i);
			break;
		}	
}

void fasta::load(const string &fn)
{	string	path(fn);
	string	tmp;
	struct stat state;
	if (path.size()>3 && path.substr(path.size()-3)==".gz")
		path.erase(path.size()-3);
	if (stat(path.c_str(),&state))
	{	path+=".gz";
		if (stat(path.c_str(),&state))
		{	throw bio_exception("fasta::load : can't open '"+fn+"'");
		}
		char	h[20];
		sprintf(h,"%d",int(getpid()));
		tmp=string("/tmp/")+h;
		string cmd(zcat+" \""+path+"\" >"+tmp);
		if (system(cmd.c_str()))
			throw bio_exception("fasta::load : error executing '"+cmd+"'");
		path=tmp;
	}
	ifstream i(path.c_str());
	i >> *this;
	if (tmp!="")
	{	string cmd(rm+" "+tmp);
		if (system(cmd.c_str()))
			throw bio_exception("fasta::load : error executing '"+cmd+"'");
	}
}

void fasta::set_startchar(unsigned char s)
{	for (list<fasta_piece>::iterator i=data.begin(); i!=data.end(); ++i)
		i->set_startchar(s);
}

void fasta::set_output_width(int x)
{	for (list<fasta_piece>::iterator i=data.begin(); i!=data.end(); ++i)
		i->set_output_width(x);
}

int fasta::count_bases() const
{	int	ret=0;
	for (list<fasta_piece>::const_iterator i=data.begin(); i!=data.end(); ++i)
		ret+=i->count_bases();
	return ret;
}	

fasta_piece& fasta::piece(int x)
{	int n=0;
	for (list<fasta_piece>::iterator i=data.begin(); i!=data.end(); ++i,++n)
		if (n==x)
			return *i;
	ostringstream o;
	o << x;
	throw bio_exception(string("fasta::piece : no such piece ")+o.str());
}

const fasta_piece& fasta::piece(int x) const
{	int n=0;
	for (list<fasta_piece>::const_iterator i=data.begin(); i!=data.end(); ++i,++n)
	{	if (n==x)
		{	return *i;
		}
	}
	ostringstream o;
	o << x;
	throw bio_exception(string("fasta::piece : no such piece ")+o.str());
}	

fasta_piece& fasta::operator [] (int x)
{	return piece(x);
}

const fasta_piece& fasta::operator [] (int x) const
{	return piece(x);
}

const string fasta::get_data(unsigned int p) const
{	if (data.size()<=p)
		return string();
	unsigned int l=0;
	for (list<fasta_piece>::const_iterator i=data.begin(); i!=data.end(); ++i,++l)
		if (l==p)
			return i->get_data();
	return string();
}

const string fasta::get_data() const
{	string ret;
	for (list<fasta_piece>::const_iterator i=data.begin(); i!=data.end(); ++i)
		ret+=i->get_data();
	return ret;	
}

const list<int> fasta::get_gaps() const
{	list<int>	ret;
	if (!data.size()) 
		return ret;
	int	pos=0;	
	for (list<fasta_piece>::const_iterator i=data.begin(); i!=--data.end(); ++i)
	{	pos+=i->count_bases();
		ret.push_back(pos);
	}
	return ret;
}

const string fasta::get_name(unsigned int p) const
{	if (data.size()<=p)
		return string();
	unsigned int l=0;
	for (list<fasta_piece>::const_iterator i=data.begin(); i!=data.end(); ++i,++l)
		if (l==p)
			return i->get_name();
	return string();
}

const string fasta::get_name() const
{	string ret;
	for (list<fasta_piece>::const_iterator i=data.begin(); i!=data.end(); ++i)
		ret+=i->get_name()+" ";
	if (ret.size())
	{	ret.erase(ret.size()-1);
	}
	return ret;	
}

bool fasta::blastn_valid() const
{	for (list<fasta_piece>::const_iterator i=data.begin(); i!=data.end(); ++i)
		if (!i->blastn_valid())
			return false;
	return true;
}

void fasta::toupper()
{	for (list<fasta_piece>::iterator i=data.begin(); i!=data.end(); ++i)
		i->toupper();
}

void fasta::tolower()
{	for (list<fasta_piece>::iterator i=data.begin(); i!=data.end(); ++i)
		i->tolower();
}

void fasta::reverse()
{	for (list<fasta_piece>::iterator i=data.begin(); i!=data.end(); ++i)
		i->reverse();
}

void fasta::complement()
{	for (list<fasta_piece>::iterator i=data.begin(); i!=data.end(); ++i)
		i->complement();
}		

void fasta::revcomp()
{	for (list<fasta_piece>::iterator i=data.begin(); i!=data.end(); ++i)
		i->revcomp();
}		

map<char,int> fasta::histogram(bool ignore_case) const
{	map<char,int>	ret;
	for (list<fasta_piece>::const_iterator i=data.begin(); i!=data.end(); ++i)
	{	for (unsigned int l=0; l<i->data.length(); ++l)
			if (ignore_case)
				ret[::toupper(i->data[l])]++;
			else
				ret[i->data[l]]++;
	}
	return ret;
}

list<search_hit> fasta::search(const string &s) const
{	int pc=0;
	list<search_hit>	ret;
	for (list<fasta_piece>::const_iterator i=data.begin(); i!=data.end(); ++i)
	{	list<int> h(i->search(s));
		for (list<int>::iterator j=h.begin(); j!=h.end(); ++j)
			ret.push_back(search_hit(pc,*j));
		pc++;	
	}
	return ret;
}	

static string fasta_rtrim(const string &t)
{       string s(t);
	while (s.size() && s[s.size()-1]==' ')
		s.erase(s.size()-1);
	return s;
}

int fasta::find_by_name(const string &s) const
{	int	ret=0;
	for (list<fasta_piece>::const_iterator i=data.begin(); i!=data.end(); ++i)
	{	if (fasta_rtrim(s)==fasta_rtrim(i->get_name()))
			return ret;
		++ret;
	}
	throw bio_exception("fasta::find_by_name : nothing found for '"+s+"'");
}	

int fasta::find_by_name_until_first_space(const std::string &s) const
{	int	ret=0;
	int	found=-1;
	for (list<fasta_piece>::const_iterator i=data.begin(); i!=data.end(); ++i)
	{	string t(i->get_name());
		if (t.find(' ')!=string::npos)
			t.erase(t.find(' '));
		if (fasta_rtrim(s)==t)
		{	if (found==-1)
			{	found=ret;
			}
			else
			{	throw bio_exception("fasta::find_by_name_until_first_space : '"+s+"' is not unique");
			}
		}
		++ret;
	}
	if (found!=-1)
		return found;
	throw bio_exception("fasta::find_by_name : nothing found for '"+s+"'");
}

istream& operator >> (istream &i,fasta &f)
{	while (!i.eof())
	{	fasta_piece	p;
		i >> p;
		f.data.push_back(p);
	}
	return i;
}	

ostream& operator << (ostream &o,const fasta &f)
{	for (list<fasta_piece>::const_iterator i=f.data.begin(); i!=f.data.end(); ++i)
		o << *i;
	return o;
}		

/* 	'*' used as stop
	assuming whole string upper or lowercase
*/

string translate(const string &s_,int start)
{	string	s(s_);
	string ret;
	size_t	pos=start;
	bool	upper=false;
	if (s.size() && s[0]==toupper(s[0]))
	{	upper=true;
		for (unsigned int l=0; l<s.size(); ++l)
			s[l]=tolower(s[l]);
	}
	while (pos+3<=s.size())
	{	switch (s[pos])
		{	case 't' :
				switch (s[pos+1])
				{	case 't' :
						switch (s[pos+2])
						{	case 't' : ret+='f'; break;	
							case 'c' : ret+='f'; break;
							case 'a' : ret+='l'; break;
							case 'g' : ret+='l'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					case 'c' :
						switch (s[pos+2])
						{	case 't' : ret+='s'; break;
							case 'c' : ret+='s'; break;
							case 'a' : ret+='s'; break;
							case 'g' : ret+='s'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					case 'a' :
						switch (s[pos+2])
						{	case 't' : ret+='y'; break;
							case 'c' : ret+='y'; break;
							case 'a' : ret+='*'; break;
							case 'g' : ret+='*'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					case 'g' :
						switch (s[pos+2])
						{	case 't' : ret+='c'; break;
							case 'c' : ret+='c'; break;
							case 'a' : ret+='*'; break;
							case 'g' : ret+='w'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					default : throw bio_exception(string("not translatable : ")+s[pos+1]);
				}	
				break;
			case 'c' :
				switch (s[pos+1])
				{	case 't' :
						switch (s[pos+2])
						{	case 't' : ret+='l'; break;
							case 'c' : ret+='l'; break;
							case 'a' : ret+='l'; break;
							case 'g' : ret+='l'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					case 'c' :
						switch (s[pos+2])
						{	case 't' : ret+='p'; break;
							case 'c' : ret+='p'; break;
							case 'a' : ret+='p'; break;
							case 'g' : ret+='p'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					case 'a' :
						switch (s[pos+2])
						{	case 't' : ret+='h'; break;
							case 'c' : ret+='h'; break;
							case 'a' : ret+='q'; break;
							case 'g' : ret+='q'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					case 'g' :
						switch (s[pos+2])
						{	case 't' : ret+='r'; break;
							case 'c' : ret+='r'; break;
							case 'a' : ret+='r'; break;
							case 'g' : ret+='r'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					default : throw bio_exception(string("not translatable : ")+s[pos+1]);
				}	
				break;
			case 'a' :
				switch (s[pos+1])
				{	case 't' :
						switch (s[pos+2])
						{	case 't' : ret+='i'; break;
							case 'c' : ret+='i'; break;
							case 'a' : ret+='i'; break;
							case 'g' : ret+='m'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					case 'c' :
						switch (s[pos+2])
						{	case 't' : ret+='t'; break;
							case 'c' : ret+='t'; break;
							case 'a' : ret+='t'; break;
							case 'g' : ret+='t'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					case 'a' :
						switch (s[pos+2])
						{	case 't' : ret+='n'; break;
							case 'c' : ret+='n'; break;
							case 'a' : ret+='k'; break;
							case 'g' : ret+='k'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					case 'g' :
						switch (s[pos+2])
						{	case 't' : ret+='s'; break;
							case 'c' : ret+='s'; break;
							case 'a' : ret+='r'; break;
							case 'g' : ret+='r'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					default : throw bio_exception(string("not translatable : ")+s[pos+1]);
				}	
				break;
			case 'g' :
				switch (s[pos+1])
				{	case 't' :
						switch (s[pos+2])
						{	case 't' : ret+='v'; break;
							case 'c' : ret+='v'; break;
							case 'a' : ret+='v'; break;
							case 'g' : ret+='v'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					case 'c' :
						switch (s[pos+2])
						{	case 't' : ret+='a'; break;
							case 'c' : ret+='a'; break;
							case 'a' : ret+='a'; break;
							case 'g' : ret+='a'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					case 'a' :
						switch (s[pos+2])
						{	case 't' : ret+='d'; break;
							case 'c' : ret+='d'; break;
							case 'a' : ret+='e'; break;
							case 'g' : ret+='e'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					case 'g' :
						switch (s[pos+2])
						{	case 't' : ret+='g'; break;
							case 'c' : ret+='g'; break;
							case 'a' : ret+='g'; break;
							case 'g' : ret+='g'; break;
							default : throw bio_exception(string("not translatable : ")+s[pos+2]);
						}	
						break;
					default : throw bio_exception(string("not translatable : ")+s[pos+1]);
				}	
				break;
			default : throw bio_exception(string("no translatable : '")+s[pos]+"'");
		}	
		pos+=3;
	}
	if (upper)
	{	for (unsigned int l=0; l<ret.size(); ++l)
			ret[l]=toupper(ret[l]);
	}
	return ret;
}	

void fasta::fast_load(const string &s)
{	clear();
	struct stat state;
	if (-1==stat(s.c_str(),&state))
	{	throw bio_exception("fasta::fast_load : can't stat '"+s+"'");
	}
	size_t	sz=state.st_size;
	int	file=open(s.c_str(),O_RDONLY);
	if (-1==file)
	{	throw bio_exception("fasta::fast_load : can't open '"+s+"'");
	}
	char *buffer=new char[sz];
	if (buffer==NULL)
	{	throw bio_exception("fasta::fast_load : can't get enough mem");
	}

	size_t	to_read_size=sz;
	char	*dst=buffer;

	do	
	{	ssize_t readed=read(file,dst,to_read_size);
		if (-1==readed)
		{	ostringstream out;
			out << "size=" << sz << " readed=" << readed;
			perror("error:");
			throw bio_exception("fasta::fast_load : error while reading '"+s+"' "+out.str());
		}
		dst+=readed;
		to_read_size-=readed;
	} while (to_read_size);	

	close(file);
	size_t pos=0;
	size_t start;
	while (pos<sz)
	{	// find header line
		if (buffer[pos]!='>')
		{	throw bio_exception("fasta::fast_load : no startchar");
		}
		pos++;
		start=pos;
		while (buffer[pos]!=0xa && buffer[pos]!=0xd)
			pos++;
		string	header(buffer+start,pos-start);
		while (buffer[pos]==0xa || buffer[pos]==0xd)
			pos++;
		string	d;
		while (pos<sz && buffer[pos]!='>')
		{	start=pos;
			while (pos<sz && buffer[pos]!=0xa && buffer[pos]!=0xd)
				pos++;
			d+=string(buffer+start,pos-start);
			while (pos<sz && (buffer[pos]==0xa || buffer[pos]==0xd))
				pos++;
		}
		data.push_back(fasta_piece());
		data.back().set_name(header);
		data.back().set_data(d);
	}
}

void get_contig_sizes(const string &infile,map<string,size_t> &result)
{	int     in=open(infile.c_str(),O_RDONLY);
	if (-1==in)
	{	throw bio_exception("get_contig_sizes: can't open '"+infile+"'");
	}
	size_t	input_size=lseek(in,0,SEEK_END);
	lseek(in,0,SEEK_SET);
	const size_t    part_size=128*1024*1024;
	char    *buffer=new char[part_size];
	size_t  pos=0;
	string 	c_name;
	size_t	c_size=0;
	bool	in_header=false;
	while (pos<input_size)
	{       ssize_t readed=read(in,buffer,part_size);
		if (readed==-1)
		{	throw bio_exception("get_contig_sizes: read failed");
		}
		for (ssize_t i=0; i<readed; ++i)
		{	if (in_header)
			{       if (buffer[i]=='\n')
				{       in_header=false;
					c_size=0;
				}
				else
				{	c_name+=buffer[i];
				}
			}
			else
			{	if (buffer[i]=='>')
				{       if (c_name!="")
					{	result[c_name]=c_size;
						c_name="";
					}
					in_header++;
				}
				else
				{	if (buffer[i]!='\n' && buffer[i]!='\t' && buffer[i]!=' ')
					{	c_size++;
					}
				}	
			}
		}
		pos+=readed;
	}	
	if (c_name!="")
	{	result[c_name]=c_size;
	}
}

