/*******************************************************************************
*                                                                              *
*    Leibniz Institute on Aging - Fritz Lipmann Institute                      *
*    Genomic Ananlysis                                                         *
*                                                                              *
*    niels jahn	                                                               *
*    2014,2015                                                                 *
*                                                                              *
*******************************************************************************/

#ifndef _FASTA_H_
#define _FASTA_H_

#include <string>
#include <fstream>
#include <list>
#include <map>
#include "bio_exception.h"

class fasta_piece
{	std::string	name;
	std::string	data;
	unsigned char		startchar;
	int		output_width;
	public:

	fasta_piece();
	fasta_piece(const fasta_piece&);
	void set_startchar(unsigned char);
	void set_output_width(int);
	const std::string& get_name() const { return name; }
	const std::string& get_data() const { return data; }
	std::string	get_part(int,int);
	void	set_name(const std::string &s) { name=s; }
	void	append_name(const std::string &s) { name+=s; }
	void	set_data(const std::string &s) { data=s; }
	void	append_data(const std::string &s) { data+=s; }
	int	count_bases() const { return data.length(); }
	int	size() const { return data.length(); }

	bool	blastn_valid() const;
	void	toupper();
	void	tolower();
	void	reverse();
	void	complement();
	void	revcomp();
	void	set(int,int,char);
	double	get_gc() const;

	std::list<int>	search(const std::string&) const;
	friend std::istream& operator >> (std::istream&,fasta_piece&);
	friend std::ostream& operator << (std::ostream&,const fasta_piece&);
	friend class fasta;
};

std::istream& operator >> (std::istream&,fasta_piece&);
std::ostream& operator << (std::ostream&,const fasta_piece&);

class search_hit
{	public:
	int	piece;
	int	pos;
	search_hit(int i,int j) { piece=i; pos=j; }
};	

class fasta
{	std::list<fasta_piece>	data;
	public:
	
	fasta();
	fasta(const fasta_piece&);
	fasta(const fasta&);
	fasta(const std::string&);

	void	load(const std::string&);
	void	fast_load(const std::string&);
	
	bool blastn_valid() const;
	std::map<char,int>	histogram(bool ignore_case=false) const;
	void	set_startchar(unsigned char);
	void	set_output_width(int);
	int	count_pieces() const { return data.size(); }
	unsigned int size() const { return data.size(); }
	int	count_bases() const;
	fasta_piece&	piece(int);
	const fasta_piece&    piece(int) const;
	fasta_piece&	operator [] (int);
	const fasta_piece&	operator [] (int) const;
	const std::string	get_data(unsigned int) const;
	const std::string	get_data() const;
	const std::list<int>	get_gaps() const;
	const std::string	get_name(unsigned int) const;
	const std::string	get_name() const;
	void	toupper();
	void	tolower();
	void	reverse();
	void	complement();
	void	revcomp();
	void	clear();
	void	add(const fasta_piece&);
	void	add_after(int pos,const fasta_piece&);
	void	erase(int);
	int	find_by_name(const std::string&) const;
	int	find_by_name_until_first_space(const std::string&) const;
	std::list<search_hit>	search(const std::string&) const;
	friend std::istream& operator >> (std::istream&,fasta&);
	friend std::ostream& operator << (std::ostream&,const fasta&);

	class iterator
	{	std::list<fasta_piece>::iterator i;
		public:
		iterator(std::list<fasta_piece>::iterator x)
		{	i=x;
		}
		fasta_piece& operator *()
		{	return *i;
		}
		fasta_piece* operator ->()
		{	return &(*i);
		}
		bool operator == (const iterator &x)
		{	return i==x.i;
		}
		bool operator != (const iterator &x)
		{	return i!=x.i;
		}
		const iterator& operator ++ ()
		{	++i;
			return *this;
		}
	};

	iterator begin()
	{	return iterator(data.begin());
	}
	iterator end()
	{	return iterator(data.end());
	}

};

std::istream& operator >> (std::istream&,fasta&);
std::ostream& operator << (std::ostream&,const fasta&);

std::string translate(const std::string&,int start=0);
void get_contig_sizes(const std::string&,std::map<std::string,size_t>&);

#endif

