/*******************************************************************************
*                                                                              *
*    Leibniz Institute on Aging - Fritz Lipmann Institute                      *
*    Genomic Ananlysis                                                         *
*                                                                              *
*    niels jahn	                                                               *
*    2014,2015                                                                 *
*                                                                              *
*******************************************************************************/

#include <string>
#include <list>
#include <iostream>
#include <set>
#include <fstream>

#include <boost/lexical_cast.hpp>

#include <sys/types.h>
#include <dirent.h>

#include "revcomp.h"
#include "search_around_class.h"

#include "cr_meth_var.h"

using namespace std;
using namespace boost;

const int version_major=0;
const int version_minor=3;

void usage(char *s)
{	cout << s << ": Usage" << endl;
	cout << "v: " << version_major << "." << version_minor << endl;
	cout <<"	-h			this help" << endl;
	cout <<	"	-f forward_primer" << endl;
	cout << "	-r reverse_primer" << endl;
	cout << "	-R revcomp(reverse_primer)" << endl;
	cout << "	-t tag 			(multiple possible)" << endl;
	cout << "	-i input 		fastq (multiple possible)" << endl;
	cout << "	-d directory		directory to search for files of type *.fastq" << endl;
	cout << "	-P pattern[:replace]	pattern or pattern:replacement" << endl;
	cout << "				(multiple possible)" << endl;
	cout << "	-b number		characters before match" << endl;
	cout << "	-a number		characters after match" << endl;
	cout << "	-n number		characters before and after match (overrides -a and -b)" << endl;
	cout << "	-H number		how much read length histogram should be shown" << endl;
	cout << "				default 5" << endl;
	cout << "				to list all use -H 0" << endl;
	cout << "	-S number		how many hits for search around should be shown" << endl;
	cout << "				default 10";
	cout << "				to list all use -S 0" << endl;
	cout << "	-s			do not sort search results" << endl;
	cout << "				(disables -S)" << endl;
	cout << "	-o			write search around data to file(s)" << endl;
	cout << "	-C			do not cut primers from search around data" << endl;
	cout << "	-X			show pattern hit stats" << endl;
	cout << "	-M file			do methy var search (fasta)" << endl;	
	cout <<	endl;
}

void create_pattern(const string &tag,const string &forward_primer,const string &reverse_primer,
	string &p1_s,string &p1_e,string &p2_s,string &p2_e)
{
	p1_s=tag+forward_primer;
	p1_e=revcomp(reverse_primer)+revcomp(tag);
	p2_s=tag+reverse_primer;
	p2_e=revcomp(forward_primer)+revcomp(tag);
}

class search_around_remover
{	size_t	req;
	public:
	search_around_remover(size_t x)
	:req(x)
	{
	}
	bool operator () (const string &s)
	{	return s.size()!=req;
	}
};	

class histo_sorter
{	public:
	bool operator () (const pair<int,int> &a,const pair<int,int> &b)
	{	if (a.first==b.first)
		{	return a.second>b.second;
		}
		return a.first>b.first;
	}
};

int main(int argc,char **argv)
{	
	string	forward_primer;
	string	reverse_primer;
	set<string>	tags;
	string	dir;
	list<string>	todo;	// files to work on
	list<string>	patterns;	// pattern[:replacement]
	size_t		before=0;
	size_t		after=0;
	size_t		around=0;

	int	max_histo_show=5;
	int	max_search_around_hits=10;

	bool	output_sad_data=false;
	bool	show_pattern_hit_stats=false;

	bool	sort_hits=true;
	bool	cut_primers=true;

	string	methy_vars_input;

	int k;
	while (-1!=(k=getopt(argc,argv,"hf:r:t:i:d:P:b:a:n:H:S:oR:XsCM:")))
	{	switch (k)
		{	case 'h' :
				usage(argv[0]);
				return 0;
			case 'f' :
				forward_primer=string(optarg);
				break;
			case 'r' :
				reverse_primer=string(optarg);
				break;
			case 't' :
				tags.insert(string(optarg));
				break;
			case 'i' :
				todo.push_back(string(optarg));
				break;
			case 'd' :
				dir=string(optarg);
				break;
			case 'P' :
				patterns.push_back(string(optarg));
				break;
			case 'b' :
				before=lexical_cast<int>(optarg);
				break;
			case 'a' :
				after=lexical_cast<int>(optarg);
				break;
			case 'n' :
				around=lexical_cast<int>(optarg);
				break;
			case 'H' :
				max_histo_show=lexical_cast<int>(optarg);
				break;
			case 'S' :
				max_search_around_hits=lexical_cast<int>(optarg);
				break;
			case 'o' :
				output_sad_data=true;
				break;
			case 'R' :
				reverse_primer=revcomp(string(optarg));
				break;
			case 'X' :
				show_pattern_hit_stats=true;
				break;
			case 's' :
				sort_hits=false;
				break;
			case 'C' :
				cut_primers=false;
				break;
			case 'M' :
				methy_vars_input=string(optarg);
				break;
			default :
				cerr << "incorrect parameter" << endl;
				return 1;
		}
	}	

	// print command line parameter
	for (int i=0; i<argc; ++i)
	{	cout << argv[i] << ' ';
	}
	cout << endl;

	if (forward_primer=="" || reverse_primer=="" || (todo.size()==0 && dir==""))
	{	cerr << "missing input" << endl;
		if (forward_primer=="") cerr << "\tforward primer" << endl;
		if (reverse_primer=="") cerr << "\treverse primer" << endl;
		if (todo.size()==0 && dir=="") cerr << "\tno input" << endl;
		cout << endl;
		usage(argv[0]);
		return 1;
	}

	if (around)
	{	before=after=around;
	}

	if (max_histo_show<0)
	{	max_histo_show=0;
	}
	if (max_search_around_hits<0 || sort_hits==false)
	{	max_search_around_hits=0;
	}

	// check tag size
	size_t tag_size=0;
	for (set<string>::iterator i=tags.begin(); i!=tags.end(); ++i)
	{
		if (tag_size!=0 && i->size()!=tag_size)
		{	cout << "differing tag sizes; don't know how to handle" << endl;
			return 0;
		}
		else
		{	tag_size=i->size();
		}
	}	

	if (tags.size()==0)
	{	tags.insert("");
	}

	fasta	methy_patterns;
	if (methy_vars_input!="")
	{	fasta	f(methy_vars_input);
		cr_methylation_variants(f,methy_patterns);
	}	

	// search for files

	if (dir!="")
	{	/* sort found files alphabetically after files given using -i
		*/
		list<string>	todo_tmp;
		DIR *dir_searched=opendir(dir.c_str());
		if (dir_searched==NULL)
		{	cout << "can't open directory '" << dir << "'" << endl;
			return 1;
		}
		dirent	*current_file;
		while (NULL!=(current_file=readdir(dir_searched)))
		{	string fn(current_file->d_name);
			size_t pos=fn.rfind(".fastq");
			if (string::npos!=pos && fn.size()-6==pos)
			{       todo_tmp.push_back(dir+"/"+fn);
			}
		}
		closedir(dir_searched);
		todo_tmp.sort();
		todo.splice(todo.end(),todo_tmp);
	}

	// handle files
	for (list<string>::iterator i=todo.begin(); i!=todo.end(); ++i)
	{	cout << endl;
		cout << "file : " << *i << " :" << endl;
		cout << endl;

		ifstream in(i->c_str());
		if (!in || !in.is_open())
		{	cout << "ERROR: can't open '" << *i << "' skipping" << endl;
			continue;
		}
		
		map<string,map<int,int> >	matches;	// string hist len
		int count=0;
		
		// insert searched stuff, so we get info about not found tags
		for (set<string>::iterator li=tags.begin(); li!=tags.end(); ++li)
		{	string p1_s,p1_e,p2_s,p2_e;
			create_pattern(*li,forward_primer,reverse_primer,
				p1_s,p1_e,p2_s,p2_e);
			matches[p1_s+"..."+p1_e]=map<int,int>();
			matches[p2_s+"..."+p2_e]=map<int,int>();
		}

		map<string,list<string> >	search_around_data;	// inserts per tag

		string	line;
		string	dummy;
		while (true)
		{	getline(in,dummy);
			getline(in,line);
			getline(in,dummy);
			getline(in,dummy);
			
			if (in.eof())
			{	break;
			}
			count++;

			for (set<string>::iterator ti=tags.begin(); ti!=tags.end(); ++ti)
			{	
				string p1_s,p1_e,p2_s,p2_e;
				create_pattern(*ti,forward_primer,reverse_primer,
					p1_s,p1_e,p2_s,p2_e);
				
				if ((p1_e.size()+p1_s.size())>line.size())	// read too short
				{	continue;
				}

				// searching TAG+FW_RPIMER+...+revcomp(REV_PRIMER)+revcomp(TAG)
				if (line.substr(0,p1_s.size())==p1_s && line.substr(line.size()-p1_e.size())==p1_e)
				{	matches[p1_s+"..."+p1_e][line.size()]++;
					string insert;
					if (cut_primers)
					{	insert=line.substr(p1_s.size());
						insert.erase(insert.size()-p1_e.size());
					}
					else
					{	insert=line;
						insert.erase(0,ti->size());
						insert.erase(insert.size()-ti->size());
					}
					search_around_data[*ti].push_back(insert);
				}
				// searching TAG+REV_PRIMER+...+revcomp(FW_PRIMER)+revcomp(TAG)
				else if (line.substr(0,p2_s.size())==p2_s && line.substr(line.size()-p2_e.size())==p2_e)
				{	matches[p2_s+"..."+p2_e][line.size()]++;
					string insert;
					if (cut_primers)
					{	insert=line.substr(p2_s.size());
						insert.erase(insert.size()-p2_e.size());
					}
					else
					{	insert=line;
						insert.erase(0,ti->size());
						insert.erase(insert.size()-ti->size());
					}
					search_around_data[*ti].push_back(revcomp(insert));
				}
			}	
		}

		int choosen_len=0;

		// output per tag
		map<string,string>	tag_output;

		for (map<string,map<int,int> >::iterator m=matches.begin(); m!=matches.end(); ++m)
		{	ostringstream	out;
			out << m->first;

			int c=0;
			for (map<int,int>::iterator j=m->second.begin(); j!=m->second.end(); ++j)
			{	c+=j->second;
			}
			out << '\t' << c;
			out << '\t' << 100.0*c/count << " %";
			out << endl;

			list<pair<int,int> >	histo;
			for (map<int,int>::iterator j=m->second.begin(); j!=m->second.end(); ++j)
			{	histo.push_back(pair<int,int>(j->second,j->first));
			}
			histo.sort(histo_sorter());

			int histo_shown=0;
			for (list<pair<int,int> >::const_iterator j=histo.begin(); j!=histo.end(); ++j)
			{	out << '\t' << j->second;
				out << '\t' << j->first;
				out << '\t' << 100.0*j->first/c << " %";
				out << endl;
				histo_shown++;
				if (max_histo_show && histo_shown>=max_histo_show)
				{	break;
				}
			}
			if (histo.size())
			{	choosen_len=histo.begin()->second;
			}
			tag_output[m->first.substr(0,tag_size)]+=out.str();
		}
		cout << "detected len " << choosen_len << endl;

		for (set<string>::iterator tag=tags.begin(); tag!=tags.end(); ++tag)
		{	// remove anything from search_around_data which not has choosen_len
			if (!cut_primers)
			{	choosen_len+=forward_primer.size()+reverse_primer.size();
			}
			search_around_data[*tag].remove_if(search_around_remover(choosen_len-2*tag_size-forward_primer.size()-reverse_primer.size()));

			cout << endl;
			if (tag_size)
			{	cout << "tag " << *tag << " :" << endl;
			}
			cout << tag_output[*tag];

			if (output_sad_data)
			{	string out_name(*i+".sad_data_"+*tag);
				if (out_name.find('/')!=string::npos)
				{	out_name.erase(0,1+out_name.rfind('/'));
				}
				ofstream out(out_name.c_str());
				if (!out || !out.is_open())
				{	cerr << "can't open '" << out_name << "' for writing" << endl;
				}
				for (list<string>::iterator c_sad=search_around_data[*tag].begin(); c_sad!=search_around_data[*tag].end(); ++c_sad)
				{	out << *c_sad << endl;
				}
				out.close();
			}	

			// handle the patterns
			if (patterns.size() && search_around_data[*tag].size())
			{	//cout << "search around ..." << endl;
				search_around_class     sac(before,after);
				sac.use_data(search_around_data[*tag]);

				for (list<string>::iterator j=patterns.begin(); j!=patterns.end(); ++j)
				{       cout << "pattern " << *j << " : " << endl;
					sac.search(*j);
					sac.output(cout,show_pattern_hit_stats,max_search_around_hits,sort_hits);
				}
			}	

			// handle methy search
			if (methy_vars_input!="" && search_around_data[*tag].size())
			{	cout << endl;
				cout << "methy patterns search '" << methy_vars_input << "'" << endl;
				map<string,pair<int,int> >      results;

				for (fasta::iterator mp=methy_patterns.begin(); mp!=methy_patterns.end(); ++mp)
				{	results[mp->get_name()].first=0;	// for showing zero results too
					for (list<string>::iterator c_sad=search_around_data[*tag].begin(); c_sad!=search_around_data[*tag].end(); ++c_sad)
					{	if (c_sad->find(mp->get_data())!=string::npos)
						{	results[mp->get_name()].first++;
						}
/*						fasta_piece fp(*mp);
						fp.revcomp();
						if (c_sad->find(fp.get_data())!=string::npos)
						{	results[mp->get_name()].second++;
						}*/
					}
					
				}
				for (map<string,pair<int,int> >::iterator mres=results.begin(); mres!=results.end(); ++mres)
				{	cout << mres->first << '\t' << mres->second.first << '\t';
//					cout << mres->first << "_revcomp\t" << mres->second.second;
					cout << endl;
				}
			}

		}	
	}

	return 0;
}	

